from __future__ import annotations

import os
import traceback
from concurrent.futures import ThreadPoolExecutor
from queue import Queue
from re import fullmatch
from subprocess import Popen, TimeoutExpired
from threading import Event, Lock, RLock, Thread, Timer
from time import monotonic, sleep
from typing import Dict, List, Optional, Sequence, Type

import grpc
from eventsourcing.application import (
    Application,
    NotificationLog,
    RecordingEvent,
    Section,
    TApplication,
)
from eventsourcing.system import (
    EventProcessingError,
    Follower,
    Leader,
    RecordingEventReceiver,
)
from eventsourcing.utils import EnvType, get_topic
from grpc import ServicerContext, local_server_credentials

from eventsourcing_grpc.application_client import (
    ClientClosedError,
    DeadlineExceeded,
    GrpcApplicationClient,
    GrpcError,
    ServiceUnavailable,
    create_client,
)
from eventsourcing_grpc.environment import GrpcEnvironment
from eventsourcing_grpc.protos.application_pb2 import (
    Empty,
    MethodReply,
    MethodRequest,
    Notification,
    NotificationsReply,
    NotificationsRequest,
    PromptRequest,
)
from eventsourcing_grpc.protos.application_pb2_grpc import (
    ApplicationServicer,
    add_ApplicationServicer_to_server,
)


class GRPCRemoteNotificationLog(NotificationLog):
    def __init__(self, leader_client: GrpcApplicationClient[Application]):
        self.leader_client = leader_client

    def __getitem__(self, section_id: str) -> Section:
        raise NotImplementedError()

    def select(
        self,
        start: int,
        limit: int,
        stop: Optional[int] = None,
        topics: Sequence[str] = (),
    ) -> List[Notification]:
        return self.leader_client.get_notifications(
            start=start, limit=limit, stop=stop, topics=list(topics)
        )


class GRPCRecordingEventReceiver(RecordingEventReceiver):
    def __init__(
        self,
        leader_name: str,
        follower_client: GrpcApplicationClient[Application],
        min_interval: float = 0.05,  # Todo: Make this configurable.
    ):
        self.leader_name = leader_name
        self.follower_client = follower_client
        self.last_prompt = monotonic()
        self.timer_lock = RLock()
        self.timer: Optional[Timer] = None
        self.min_interval = min_interval

    def receive_recording_event(self, recording_event: RecordingEvent) -> None:
        with self.timer_lock:
            time_since_last_prompt = monotonic() - self.last_prompt
            if time_since_last_prompt < self.min_interval:
                wait_to_prompt = self.min_interval - time_since_last_prompt
                # print("Wait to prompt:", wait_to_prompt)
                if self.timer is None:
                    self.start_timer(wait_to_prompt)
            else:
                if self.timer is None:
                    self.prompt_follower()

    def start_timer(self, wait_to_prompt: float) -> None:
        timer = Timer(interval=wait_to_prompt, function=self.prompt_follower)
        timer.daemon = False
        timer.start()
        self.timer = timer

    def prompt_follower(self) -> None:
        with self.timer_lock:
            try:
                self.follower_client.prompt(name=self.leader_name)
            except ServiceUnavailable:
                print(
                    f"Error: Prompting of {self.follower_client.application_name} by",
                    f"{self.follower_client.owner_name} failed (service unavailable)",
                )
            except DeadlineExceeded:
                print(
                    f"Error: Prompting of {self.follower_client.application_name} by",
                    f"{self.follower_client.owner_name} failed (deadline exceeded)",
                )
            except (GrpcError, ValueError, AttributeError, ClientClosedError):
                # Todo: Log failure to prompt downstream application,
                #  possibly the connection is already closed.
                pass
            finally:
                self.last_prompt = monotonic()
                self.stop_timer()

    def stop_timer(self) -> None:
        if self.timer:
            self.timer.cancel()
            self.timer = None

    def __del__(self) -> None:
        with self.timer_lock:
            self.stop_timer()


class ApplicationService(ApplicationServicer):
    def __init__(
        self, application: Application, max_pull_interval: float, is_stopping: Event
    ) -> None:
        self.application = application
        self.max_pull_interval = max_pull_interval
        self.is_stopping = is_stopping
        self.transcoder = application.construct_transcoder()
        self.has_started = Event()
        self.prompted_names: List[str] = []
        self.prompted_names_lock = Lock()
        self.is_prompted = Event()
        self.last_prompt_times: Dict[str, float] = {}
        self.self_prompt_loop_has_stopped = Event()

    def Ping(self, request: Empty, context: ServicerContext) -> Empty:
        return Empty()

    def CallApplicationMethod(
        self, request: MethodRequest, context: ServicerContext
    ) -> MethodReply:
        method_name = request.method_name
        args = self.transcoder.decode(request.args)
        kwargs = self.transcoder.decode(request.kwargs)
        method = getattr(self.application, method_name)
        response = method(*args, **kwargs)
        reply = MethodReply()
        reply.data = self.transcoder.encode(response)
        return reply

    def GetNotifications(
        self, request: NotificationsRequest, context: ServicerContext
    ) -> NotificationsReply:
        start = int(request.start)
        limit = int(request.limit)
        stop = int(request.stop) if request.stop else None
        topics = request.topics
        notifications = self.application.notification_log.select(
            start=start, limit=limit, stop=stop, topics=topics
        )
        return NotificationsReply(
            notifications=[
                Notification(
                    id=str(n.id),
                    originator_id=n.originator_id.hex,
                    originator_version=str(n.originator_version),
                    topic=n.topic,
                    state=n.state,
                )
                for n in notifications
            ]
        )

    def Prompt(self, request: PromptRequest, context: ServicerContext) -> Empty:
        leader_name = request.upstream_name
        self.prompt(leader_name)
        return Empty()

    def prompt(self, leader_name: str) -> None:
        with self.prompted_names_lock:
            if leader_name not in self.prompted_names:
                self.prompted_names.append(leader_name)
                self.last_prompt_times[leader_name] = monotonic()
                self.is_prompted.set()

    def self_prompt_loop(self) -> None:
        try:
            if self.max_pull_interval > 0 and isinstance(self.application, Follower):
                wait_timeout = self.max_pull_interval
                while not self.is_stopping.wait(timeout=wait_timeout):
                    wait_timeout = self.max_pull_interval
                    for leader_name in self.application.readers:
                        last_time = self.last_prompt_times.get(leader_name, 0)
                        time_remaining = (
                            last_time + self.max_pull_interval - monotonic()
                        )
                        if (
                            time_remaining < 0.1
                        ):  # wait(timeout) sometimes returns early
                            self.prompt(leader_name)
                        else:
                            wait_timeout = min(time_remaining, wait_timeout)
        except BaseException:
            if self.is_stopping.is_set():
                pass
            else:
                raise
        finally:
            self.self_prompt_loop_has_stopped.set()

    def stop(self) -> None:
        self.is_stopping.set()
        self.is_prompted.set()
        self.self_prompt_loop_has_stopped.wait()


class GrpcApplicationServer:
    def __init__(
        self,
        app_class: Type[Application],
        env: EnvType,
        is_stopping: Optional[Event] = None,
        max_workers: Optional[int] = None,
    ) -> None:
        self.max_workers = max_workers
        self.lock = Lock()
        self.is_starting = Event()
        self.has_started = Event()
        self.is_stopping = is_stopping or Event()
        self.init_lead_and_follow_has_stopped = Event()
        self.pull_and_process_loop_has_stopped = Event()
        self.has_stopped = Event()
        self.grpc_env = GrpcEnvironment(env=env)
        self.application = app_class(env=env)
        self.address = self.grpc_env.get_server_address(self.application.name)
        if fullmatch("localhost:[0-9]+", self.address):
            self.server_credentials = local_server_credentials()
        else:
            ssl_private_key_path = self.grpc_env.get_ssl_private_key_path()
            ssl_certificate_path = self.grpc_env.get_ssl_certificate_path()
            ssl_root_certificate_path = self.grpc_env.get_ssl_root_certificate_path()
            if ssl_private_key_path is None:
                raise ValueError("SSL server private key path not given")
            if ssl_certificate_path is None:
                raise ValueError("SSL server certificate path not given")
            if ssl_root_certificate_path is None:
                raise ValueError("SSL root certificate path not given")
            with open(ssl_private_key_path, "rb") as f:
                ssl_private_key = f.read()
            with open(ssl_certificate_path, "rb") as f:
                ssl_certificate = f.read()
            with open(ssl_root_certificate_path, "rb") as f:
                ssl_root_certificate = f.read()

            # create server credentials
            private_key_certificate_chain_pairs = ((ssl_private_key, ssl_certificate),)
            self.server_credentials = grpc.ssl_server_credentials(
                private_key_certificate_chain_pairs=private_key_certificate_chain_pairs,
                root_certificates=ssl_root_certificate,
                require_client_auth=True,
            )

        self.clients_lock = Lock()
        self.clients: Dict[str, GrpcApplicationClient[Application]] = {}
        self.max_pull_interval = self.grpc_env.get_max_pull_interval(
            self.application.name
        )
        self.maximum_concurrent_rpcs = None
        self.compression = None

    def start(self) -> None:
        """
        Starts gRPC server.
        """
        with self.lock:
            if self.has_started.is_set() or self.is_starting.is_set():
                # print(self.application.name, "already running")
                return
            self.has_stopped.clear()
            self.init_lead_and_follow_has_stopped.clear()
            self.is_starting.set()
        try:
            # print("Starting application server:", self.application.name)
            # print(self.application.name, "starting threadpool executor")
            self.executor: ThreadPoolExecutor = ThreadPoolExecutor(
                max_workers=self.max_workers
            )
            self.executor.submit(self.stop_on_is_stopping)
            # print(self.application.name, "started threadpool executor")
            # print(self.application.name, "starting grpc server")
            self.grpc_server = grpc.server(
                thread_pool=self.executor,
                maximum_concurrent_rpcs=self.maximum_concurrent_rpcs,
                compression=self.compression,
            )
            # print(self.application.name, "creating application service")
            self.service = ApplicationService(
                self.application,
                max_pull_interval=self.max_pull_interval,
                is_stopping=self.is_stopping,
            )
            add_ApplicationServicer_to_server(self.service, self.grpc_server)
            # print(self.application.name, "adding secure port")
            self.grpc_server.add_secure_port(self.address, self.server_credentials)
            # print(self.application.name, "starting grpc server")
            self.grpc_server.start()

            # print(self.application.name, "submitting jobs to executor")
            self.executor.submit(self.init_lead_and_follow)
            self.executor.submit(self.pull_and_process_loop)
            self.executor.submit(self.service.self_prompt_loop)
            print(self.application.name, "server started:", self.address)
        finally:
            with self.lock:
                self.has_started.set()
                self.is_starting.clear()

    def stop_on_is_stopping(self) -> None:
        self.has_started.wait()
        self.is_stopping.wait()
        self._stop()

    def init_lead_and_follow(self) -> None:
        try:
            system = self.grpc_env.get_system()
            if system is not None:
                followers = system.leads[self.application.name]
                # print(self.application.name, "leads:", ", ".join(followers) )
                leaders = system.leads[self.application.name]
                # print(self.application.name, "follows:", ", ".join(leaders) )

                for follower_name in followers:
                    follower_cls = system.follower_cls(follower_name)
                    follower_client = self.get_client(
                        owner_name=self.application.name,
                        app_class=follower_cls,
                        env=self.grpc_env.env,
                    )
                    recording_event_receiver = GRPCRecordingEventReceiver(
                        leader_name=self.application.name,
                        follower_client=follower_client,
                    )
                    assert isinstance(self.application, Leader)
                    self.application.lead(follower=recording_event_receiver)

                for leader_name in leaders:
                    leader_cls = system.follower_cls(leader_name)
                    leader_client = self.get_client(
                        owner_name=self.application.name,
                        app_class=leader_cls,
                        env=self.grpc_env.env,
                    )
                    notification_log = GRPCRemoteNotificationLog(
                        leader_client=leader_client
                    )
                    assert isinstance(self.application, Follower)
                    self.application.follow(name=leader_name, log=notification_log)
                    # Prompt to catch up on anything new.
                    self.service.prompt(leader_name)

        except BaseException as e:
            if self.is_stopping.is_set():
                pass
            else:
                print(traceback.format_exc())
                print("Error initialising lead and follow connections:", e)
                raise e
        finally:
            self.init_lead_and_follow_has_stopped.set()

    def pull_and_process_loop(self) -> None:
        try:
            while not self.is_stopping.is_set():
                # Keep checking if is stopped.
                if not self.service.is_prompted.wait(timeout=1):
                    continue

                with self.service.prompted_names_lock:
                    prompted_names = self.service.prompted_names
                    self.service.prompted_names = []
                    self.service.is_prompted.clear()

                for name in prompted_names:
                    if self.is_stopping.is_set():
                        break
                    try:
                        assert isinstance(self.application, Follower)
                        if name not in self.application.readers:
                            # print(f"{self.application.name} can't pull
                            # from {name} without reader")
                            self.service.prompt(name)
                            sleep(1)
                        else:
                            print(f"{self.application.name} pulling from {name}...")
                            self.application.pull_and_process(name)
                    except ServiceUnavailable:
                        print(f"Error: {name} pulling failed (service unavailable)")
                    except DeadlineExceeded:
                        print(f"Error: {name} pulling failed (deadline exceeded)")
                    except Exception as e:
                        if self.is_stopping.is_set():
                            raise
                        error = EventProcessingError(str(e))
                        error.__cause__ = e
                        # Todo: Log the error.
                        print(
                            f"General error in {self.application.name} when",
                            "pulling and processing {name} events:",
                            traceback.format_exc(),
                        )
                        self.service.prompt(name)
                        # print("Sleeping after error....")
                        sleep(1)
        except BaseException as e:
            if self.is_stopping.is_set():
                pass
            else:
                print(traceback.format_exc())
                print("Error in pull_and_process loop:", e)
                raise
        finally:
            print("Pull and process loop exited......")
            self.pull_and_process_loop_has_stopped.set()

    def get_client(
        self, owner_name: str, app_class: Type[TApplication], env: EnvType
    ) -> GrpcApplicationClient[Application]:
        with self.clients_lock:
            if self.is_stopping.is_set() or self.has_stopped.is_set():
                raise AssertionError("Server has been stopped already")

            try:
                client = self.clients[app_class.name]
            except KeyError:
                client = create_client(
                    owner_name=owner_name,
                    app_class=app_class,
                    env=env,
                    is_stopping=self.is_stopping,
                )
                client.connect()
                if self.is_stopping.is_set() or self.has_stopped.is_set():
                    client.close()
                    client.assert_client_not_closed()
                else:
                    self.clients[app_class.name] = client
            return client

    def wait_for_termination(self) -> None:
        if not self.is_stopping.is_set():
            self.grpc_server.wait_for_termination()

    def stop(self, wait: bool = False) -> None:
        self.is_stopping.set()
        if wait:
            self.has_stopped.wait()

    def _stop(self) -> None:
        try:
            self.service.self_prompt_loop_has_stopped.wait()
            self.pull_and_process_loop_has_stopped.wait()
            self.init_lead_and_follow_has_stopped.wait()
            if not self.has_started.is_set():
                # print("stop() is_running not set, returning")
                return

            self.grpc_server.stop(grace=5)

            # print("stop() calling stop() on service...")
            self.service.stop()
            # print("stop() called stop() on service")
            # print("stop() shutting down executor")
            # Todo: Wait or not wait? that is the question (here).
            # self.executor.shutdown(wait=True, cancel_futures=True)  # type: ignore
            # print("stop() shut down executor")
            # print("stop() closing clients...")
            for client in self.clients.values():
                client.close()
            # print("stop() closed clients")

        finally:
            print("Stopped application server", self.application.name)

            with self.lock:
                self.has_started.clear()
                self.is_stopping.clear()
                self.has_stopped.set()

    def __del__(self) -> None:
        if hasattr(self, "has_stopped") and not self.has_stopped.is_set():
            self.stop()


def start_server(
    app_class: Type[Application],
    env: Optional[EnvType] = None,
    is_stopping: Optional[Event] = None,
) -> GrpcApplicationServer:
    server = GrpcApplicationServer(
        app_class=app_class,
        env=env or os.environ.copy(),
        is_stopping=is_stopping,
    )
    server.start()
    return server


class ServerSubprocess(Thread):
    def __init__(
        self, app_class: Type[Application], env: EnvType, has_errored: Event
    ) -> None:
        super().__init__(daemon=True)
        self.app_class = app_class
        self.env = env
        self.has_errored = has_errored
        self.has_started = Event()
        self.is_terminating = Event()
        self.has_stopped = Event()
        self.proc: Optional[Popen[bytes]] = None
        self.command_queue: "Queue[str]" = Queue()
        self.command_thread = Thread(target=self.command_loop, daemon=True)
        self.command_lock = Lock()

    def command_loop(self) -> None:
        while True:
            command = self.command_queue.get()
            if command == "TERMINATE":
                with self.command_lock:
                    if self.is_terminating.is_set():
                        return
                    else:
                        self.is_terminating.set()

                if self.proc is not None and self.proc.poll() is None:
                    self.proc.terminate()
                    try:
                        self.proc.wait(timeout=10)
                    except TimeoutExpired:
                        print("Timed out waiting for process to terminate. Killing....")
                        self.proc.kill()
                        self.proc.wait(timeout=1)
                        # print("Processor exit code: %s" % process.poll())
                self.has_stopped.set()
                break
            else:
                raise NotImplementedError(f"Command not supported: {command}")

    def run(self) -> None:
        self.command_thread.start()
        try:
            env = dict(self.env)
            env["APPLICATION_TOPIC"] = get_topic(self.app_class)
            self.proc = Popen(
                ["eventsourcing_grpc_server"],
                close_fds=True,
                env=env,
            )
        except BaseException:
            self.stop()
            self.has_errored.set()
            self.has_started.set()
            raise
        else:
            self.has_started.set()
            if self.proc.wait():
                self.stop()
                self.has_errored.set()
        # print("Subprocess exited, status code", self.process.returncode)

    def stop(self) -> None:
        """
        Stops given gRPC process.
        """
        self.command_queue.put("TERMINATE")


def start_server_subprocess(
    app_class: Type[Application], env: EnvType, has_errored: Event
) -> ServerSubprocess:
    thread = ServerSubprocess(app_class, env, has_errored)
    thread.start()
    return thread



import os
from functools import wraps
from re import fullmatch
from threading import Event, Lock
from typing import Any, Generic, List, Optional, Sequence, Type, cast
from uuid import UUID

import grpc
from eventsourcing.application import NotificationLog, Section, TApplication
from eventsourcing.persistence import Notification, Transcoder
from eventsourcing.utils import EnvType, retry
from grpc import (
    Channel,
    ChannelConnectivity,
    StatusCode,
    local_channel_credentials,
    ssl_channel_credentials,
)
from grpc._channel import _InactiveRpcError

from eventsourcing_grpc.environment import GrpcEnvironment
from eventsourcing_grpc.protos.application_pb2 import (
    Empty,
    MethodRequest,
    NotificationsReply,
    NotificationsRequest,
    PromptRequest,
)
from eventsourcing_grpc.protos.application_pb2_grpc import ApplicationStub


class GrpcError(Exception):
    pass


class ServiceUnavailable(GrpcError):
    pass


class DeadlineExceeded(GrpcError):
    pass


def errors(f: Any) -> Any:
    @wraps(f)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return f(*args, **kwargs)
        except _InactiveRpcError as e:
            # print(f"Error from {args[0].client_name} to {args[0].address}
            # calling func:", f, repr(e))

            if isinstance(e, _InactiveRpcError):
                if e._state.code == StatusCode.UNAVAILABLE:
                    raise ServiceUnavailable(repr(e)) from None
                elif e._state.code == StatusCode.DEADLINE_EXCEEDED:
                    raise DeadlineExceeded(repr(e)) from None
            raise GrpcError(repr(e)) from None

    return wrapper


class ClientClosedError(GrpcError):
    pass


class ClientNotReady(GrpcError):
    pass


class GrpcApplicationClient(Generic[TApplication]):
    def __init__(
        self,
        application_name: str,
        owner_name: str,
        address: str,
        transcoder: Transcoder,
        request_deadline: int = 5,
        ssl_root_certificate_path: Optional[str] = None,
        ssl_private_key_path: Optional[str] = None,
        ssl_certificate_path: Optional[str] = None,
        is_stopping: Optional[Event] = None,
    ) -> None:
        self.application_name = application_name
        self.owner_name = owner_name or "Client"
        self.address = address
        self.is_stopping = is_stopping or Event()

        if fullmatch("localhost:[0-9]+", self.address):
            self.credentials = local_channel_credentials()
        else:
            if ssl_root_certificate_path is None:
                raise ValueError("SSL root certificate path not given")
            if ssl_private_key_path is None:
                raise ValueError("SSL client private key path not given")
            if ssl_certificate_path is None:
                raise ValueError("SSL client certificate path not given")
            with open(ssl_root_certificate_path, "rb") as f:
                ssl_root_certificate = f.read()
            with open(ssl_private_key_path, "rb") as f:
                ssl_private_key = f.read()
            with open(ssl_certificate_path, "rb") as f:
                ssl_certificate = f.read()
            self.credentials = ssl_channel_credentials(
                root_certificates=ssl_root_certificate,
                private_key=ssl_private_key,
                certificate_chain=ssl_certificate,
            )

        self.transcoder = transcoder
        self.channel: Optional[Channel] = None
        self.request_deadline = request_deadline
        self.is_closed = False
        self.reconnect_lock = Lock()

    @property
    def app(self) -> TApplication:
        return cast(TApplication, ApplicationProxy(self))

    def connect(self, max_attempts: int = 0) -> None:
        """
        Connect client to server at given address.
        """
        # attempts = 0
        # start = time()
        # while True:
        if self.is_stopping and self.is_stopping.is_set():
            return
        # attempts += 1
        # self._close()
        self.channel = grpc.secure_channel(
            self.address,
            credentials=self.credentials,
            options=[
                # ("grpc.initial_reconnect_backoff_ms", 1000),
                ("grpc.max_reconnect_backoff_ms", 5000),
                # ("grpc.min_reconnect_backoff_ms", 5000),
                # ("grpc.max_connection_age_ms", 5000),  # seems to have no effect
                # ("grpc.max_connection_age_grace_ms", 5000),  # seems to have no effect
                # ("grpc.client_idle_timeout_ms", 15000),
                # ("grpc.max_connection_idle_ms", 15000),
            ],
        )
        self.stub = ApplicationStub(self.channel)
        self.channel.subscribe(self.handle_channel_state_change)

        # future = channel_ready_future(self.channel)
        # connect_deadline = min(0.5 * attempts, self.request_deadline)
        # try:
        #     future.result(timeout=connect_deadline)
        # except FutureTimeoutError:
        #     if max_attempts - attempts == 0:
        #         print(
        #             f"{self.owner_name} failed to connect to",
        #             f"{self.address}",
        #             f"after {(time() - start):.2f}s",
        #             f"(attempt {attempts}).",
        #         )
        #         raise ChannelConnectTimeout(self.address) from None
        #     else:
        #         print(
        #             f"{self.owner_name} timed out connecting to",
        #             f"{self.address}",
        #             f"after {(time() - start):.2f}s",
        #             f"(attempt {attempts}).",
        #             f"is_stopping event status: {self.is_stopping.is_set()}",
        #             "Retrying...",
        #         )
        #         continue
        # print(
        #     f"{self.owner_name} connected successfully to",
        #     f"{self.address}",
        #     f"after {(time() - start):.2f}s",
        #     f"(attempt {attempts})",
        # )
        # break

    def handle_channel_state_change(
        self, channel_connectivity: ChannelConnectivity
    ) -> None:
        print("Channel state change:", channel_connectivity)
        # if channel_connectivity == ChannelConnectivity.TRANSIENT_FAILURE:
        # try:
        #     self.reconnect_lock.acquire()
        # print("Reconnecting")
        # i
        # sleep(10)

    def __del__(self) -> None:
        self.close()

    def close(self) -> None:
        """
        Closes the client's GPRC channel.
        """
        self.is_closed = True
        self._close()

    def _close(self) -> None:
        if hasattr(self, "channel") and self.channel is not None:
            # print("closing channel...")
            self.channel.close()
            # print("closed channel")
            self.channel = None
        # Todo: Deal with calls when stub is None.
        # if hasattr(self, "stub"):
        #     self.stub = None

    @errors
    def ping(self) -> None:
        """
        Sends a Ping request to the server.
        """
        self.assert_client_not_closed()
        self.stub.Ping(Empty(), timeout=self.request_deadline)

    @errors
    def prompt(self, name: str) -> None:
        self.assert_client_not_closed()
        self.stub.Prompt(
            PromptRequest(upstream_name=name), timeout=self.request_deadline
        )

    @errors
    def get_notifications(
        self,
        start: int,
        limit: int,
        stop: Optional[int] = None,
        topics: Sequence[str] = (),
    ) -> List[Notification]:
        self.assert_client_not_closed()
        request = NotificationsRequest(
            start=str(start),
            limit=str(limit),
            stop="" if stop is None else str(stop),
            topics=topics,
        )
        notifications_reply = self.stub.GetNotifications(
            request, timeout=self.request_deadline
        )
        assert isinstance(notifications_reply, NotificationsReply)
        return [
            Notification(
                id=int(n.id),
                originator_id=UUID(n.originator_id),
                originator_version=int(n.originator_version),
                topic=n.topic,
                state=n.state,
            )
            for n in notifications_reply.notifications
        ]

    @retry(
        (DeadlineExceeded, ServiceUnavailable, ClientNotReady), max_attempts=10, wait=1
    )
    @errors
    def call_application_method(self, method_name: str, args: Any, kwargs: Any) -> Any:
        self.assert_client_not_closed()
        request = MethodRequest(
            method_name=method_name,
            args=self.transcoder.encode(args),
            kwargs=self.transcoder.encode(kwargs),
        )
        response = self.stub.CallApplicationMethod(
            request, timeout=self.request_deadline
        )
        return self.transcoder.decode(response.data)

    def assert_client_not_closed(self) -> None:
        if self.is_closed:
            raise ClientClosedError("Client is closed")


class ApplicationMethodProxy:
    def __init__(self, client: GrpcApplicationClient[TApplication], method_name: str):
        self.client = client
        self.method_name = method_name

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """
        Calls named method on server's application with given args.
        """
        return self.client.call_application_method(self.method_name, args, kwargs)


class ApplicationProxy:
    def __init__(self, client: GrpcApplicationClient[TApplication]) -> None:
        self.client = client

    @property
    def notification_log(self) -> NotificationLog:
        return NotificationLogProxy(self)

    def __getattr__(self, item: str) -> ApplicationMethodProxy:
        return ApplicationMethodProxy(self.client, item)


class NotificationLogProxy(NotificationLog):
    def __init__(self, application_proxy: ApplicationProxy):
        self.application_proxy = application_proxy

    def __getitem__(self, section_id: str) -> Section:
        raise NotImplementedError()

    def select(
        self,
        start: int,
        limit: int,
        stop: Optional[int] = None,
        topics: Sequence[str] = (),
    ) -> List[Notification]:
        return self.application_proxy.client.get_notifications(
            start=start, limit=limit, stop=stop, topics=topics
        )


def create_client(
    app_class: Type[TApplication],
    env: EnvType,
    owner_name: str,
    is_stopping: Optional[Event] = None,
) -> GrpcApplicationClient[TApplication]:

    application = app_class(env=env)
    grpc_env = GrpcEnvironment(env=env)
    client: GrpcApplicationClient[TApplication] = GrpcApplicationClient(
        application_name=application.name,
        owner_name=owner_name,
        address=grpc_env.get_server_address(app_class.name),
        transcoder=application.construct_transcoder(),
        ssl_root_certificate_path=grpc_env.get_ssl_root_certificate_path(),
        ssl_private_key_path=grpc_env.get_ssl_private_key_path(),
        ssl_certificate_path=grpc_env.get_ssl_certificate_path(),
        is_stopping=is_stopping,
    )
    return client


def connect(
    app_class: Type[TApplication],
    env: Optional[EnvType] = None,
    owner_name: str = "",
    max_attempts: int = 0,
) -> GrpcApplicationClient[TApplication]:
    client = create_client(
        app_class=app_class,
        env=env or os.environ.copy(),
        owner_name=owner_name,
    )
    client.connect(max_attempts=max_attempts)
    return client



from typing import Any, Dict, Optional, cast
from uuid import NAMESPACE_URL, UUID, uuid5

from eventsourcing.application import ProcessingEvent
from eventsourcing.domain import Aggregate, DomainEventProtocol, event
from eventsourcing.persistence import Transcoder, Transcoding
from eventsourcing.system import ProcessApplication, System


class Order(Aggregate):
    def __init__(self) -> None:
        self.is_reserved = False
        self.is_paid = False
        self.reservation_id: Optional[UUID] = None
        self.payment_id: Optional[UUID] = None

    class Reserved(Aggregate.Event):
        reservation_id: UUID

    @event(Reserved)
    def set_is_reserved(self, reservation_id: UUID) -> None:
        assert not self.is_reserved, "Order {} already reserved.".format(self.id)
        self.is_reserved = True
        self.reservation_id = reservation_id

    @event("Paid")
    def set_is_paid(self, payment_id: UUID) -> None:
        assert not self.is_paid, "Order {} already paid.".format(self.id)
        self.is_paid = True
        self.payment_id = payment_id


class Reservation(Aggregate):
    class Created(Aggregate.Created):
        order_id: UUID

    @event(Created)
    def __init__(self, order_id: UUID) -> None:
        self.order_id = order_id

    @classmethod
    def create_id(cls, order_id: UUID) -> UUID:
        return uuid5(NAMESPACE_URL, f"/reservations/{order_id}/")


class Payment(Aggregate):
    class Created(Aggregate.Created):
        order_id: UUID

    @event(Created)
    def __init__(self, order_id: UUID):
        self.order_id = order_id

    @classmethod
    def create_id(cls, order_id: UUID) -> UUID:
        return uuid5(NAMESPACE_URL, f"/payments/{order_id}/")


class OrderAsDict(Transcoding):
    type = Order
    name = "order_as_dict"

    def encode(self, obj: Order) -> Dict[str, Any]:
        return obj.__dict__

    def decode(self, data: Dict[str, Any]) -> Order:
        aggregate = object.__new__(Order)
        aggregate.__dict__.update(data)
        return aggregate


class Orders(ProcessApplication):
    def register_transcodings(self, transcoder: Transcoder) -> None:
        super(Orders, self).register_transcodings(transcoder)
        transcoder.register(OrderAsDict())

    def create_new_order(self) -> UUID:
        order = Order()
        self.save(order)
        # print("Order was created")
        return order.id

    def get_order(self, order_id: UUID) -> Dict[str, Any]:
        order = self._get_order(order_id)
        return {
            "id": order.id,
            "is_reserved": order.is_reserved,
            "is_paid": order.is_paid,
            "reservation_id": order.reservation_id,
            "payment_id": order.payment_id,
            "created_on": order.created_on,
            "modified_on": order.modified_on,
        }

    def is_order_reserved(self, order_id: UUID) -> bool:
        order = self._get_order(order_id)
        return order is not None and order.is_reserved

    def is_order_paid(self, order_id: UUID) -> bool:
        order = self._get_order(order_id)
        return order is not None and order.is_paid

    def _get_order(self, order_id: UUID) -> Order:
        return cast(Order, self.repository.get(order_id))

    def policy(
        self,
        domain_event: DomainEventProtocol,
        processing_event: ProcessingEvent,
    ) -> None:
        if isinstance(domain_event, Reservation.Created):
            # Set the order as reserved.
            order = self._get_order(order_id=domain_event.order_id)
            assert not order.is_reserved
            order.set_is_reserved(domain_event.originator_id)
            processing_event.collect_events(order)
            # print("Order was reserved")

        elif isinstance(domain_event, Payment.Created):
            # Set the order as paid.
            order = self._get_order(domain_event.order_id)
            assert not order.is_paid
            order.set_is_paid(domain_event.originator_id)
            processing_event.collect_events(order)
            # print("Order was paid")


class Reservations(ProcessApplication):
    def policy(
        self,
        domain_event: DomainEventProtocol,
        processing_event: ProcessingEvent,
    ) -> None:
        if isinstance(domain_event, Order.Created):
            # Create a reservation.
            reservation = Reservation(order_id=domain_event.originator_id)
            processing_event.collect_events(reservation)


class Payments(ProcessApplication):
    def policy(
        self,
        domain_event: DomainEventProtocol,
        processing_event: ProcessingEvent,
    ) -> None:
        if isinstance(domain_event, Order.Reserved):
            # Make a payment.
            payment = Payment(order_id=domain_event.originator_id)
            processing_event.collect_events(payment)


system = System([[Orders, Reservations, Orders, Payments, Orders]])



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class Empty(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    def __init__(
        self,
    ) -> None: ...

global___Empty = Empty

class MethodRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    METHOD_NAME_FIELD_NUMBER: builtins.int
    ARGS_FIELD_NUMBER: builtins.int
    KWARGS_FIELD_NUMBER: builtins.int
    method_name: typing.Text
    args: builtins.bytes
    kwargs: builtins.bytes
    def __init__(
        self,
        *,
        method_name: typing.Text = ...,
        args: builtins.bytes = ...,
        kwargs: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "args", b"args", "kwargs", b"kwargs", "method_name", b"method_name"
        ],
    ) -> None: ...

global___MethodRequest = MethodRequest

class MethodReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    DATA_FIELD_NUMBER: builtins.int
    data: builtins.bytes
    def __init__(
        self,
        *,
        data: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["data", b"data"]
    ) -> None: ...

global___MethodReply = MethodReply

class NotificationsRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    START_FIELD_NUMBER: builtins.int
    LIMIT_FIELD_NUMBER: builtins.int
    STOP_FIELD_NUMBER: builtins.int
    TOPICS_FIELD_NUMBER: builtins.int
    start: typing.Text
    limit: typing.Text
    stop: typing.Text
    @property
    def topics(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[
        typing.Text
    ]: ...
    def __init__(
        self,
        *,
        start: typing.Text = ...,
        limit: typing.Text = ...,
        stop: typing.Text = ...,
        topics: typing.Optional[typing.Iterable[typing.Text]] = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "limit", b"limit", "start", b"start", "stop", b"stop", "topics", b"topics"
        ],
    ) -> None: ...

global___NotificationsRequest = NotificationsRequest

class Notification(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    ID_FIELD_NUMBER: builtins.int
    ORIGINATOR_ID_FIELD_NUMBER: builtins.int
    ORIGINATOR_VERSION_FIELD_NUMBER: builtins.int
    TOPIC_FIELD_NUMBER: builtins.int
    STATE_FIELD_NUMBER: builtins.int
    id: typing.Text
    originator_id: typing.Text
    originator_version: typing.Text
    topic: typing.Text
    state: builtins.bytes
    def __init__(
        self,
        *,
        id: typing.Text = ...,
        originator_id: typing.Text = ...,
        originator_version: typing.Text = ...,
        topic: typing.Text = ...,
        state: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "id",
            b"id",
            "originator_id",
            b"originator_id",
            "originator_version",
            b"originator_version",
            "state",
            b"state",
            "topic",
            b"topic",
        ],
    ) -> None: ...

global___Notification = Notification

class NotificationsReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    NOTIFICATIONS_FIELD_NUMBER: builtins.int
    @property
    def notifications(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        global___Notification
    ]: ...
    def __init__(
        self,
        *,
        notifications: typing.Optional[typing.Iterable[global___Notification]] = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["notifications", b"notifications"]
    ) -> None: ...

global___NotificationsReply = NotificationsReply

class PromptRequest(google.protobuf.message.Message):
    """message FollowRequest {
     string name = 1;
     string address = 2;
    }

    message LeadRequest {
     string name = 1;
     string address = 2;
    }

    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    UPSTREAM_NAME_FIELD_NUMBER: builtins.int
    upstream_name: typing.Text
    def __init__(
        self,
        *,
        upstream_name: typing.Text = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["upstream_name", b"upstream_name"]
    ) -> None: ...

global___PromptRequest = PromptRequest



from __future__ import annotations

import os
from itertools import count
from threading import Event, Lock
from typing import Dict, List, Optional, Type, cast

from eventsourcing.application import Application, TApplication
from eventsourcing.system import Runner, System
from eventsourcing.utils import EnvType

from eventsourcing_grpc.application_client import GrpcApplicationClient
from eventsourcing_grpc.application_server import (
    GrpcApplicationServer,
    ServerSubprocess,
    start_server,
    start_server_subprocess,
)
from eventsourcing_grpc.environment import GrpcEnvironment


class GrpcRunner(Runner):
    """
    System runner that uses gRPC to communicate between process applications.
    """

    def __init__(self, system: System, env: Optional[EnvType] = None):
        """
        Initialises runner with the given :class:`System`.
        """
        super().__init__(system=system, env=env)

        self.is_stopping = Event()
        self._env = os.environ.copy()
        if system.topic:
            self._env["SYSTEM_TOPIC"] = system.topic
        if env is not None:
            self._env.update(env)

        self.apps: Dict[str, Application] = {}
        self.port_generator = count(start=50051)
        self.clients: Dict[str, GrpcApplicationClient[Application]] = {}
        self.servers: Dict[str, GrpcApplicationServer] = {}
        self.server_subprocesses: List[ServerSubprocess] = []
        self.lock = Lock()
        self.has_errored = Event()
        self.has_stopped = Event()
        self.has_started = Event()

    def start(self, with_subprocesses: bool = False) -> None:
        super().start()

        try:
            # Construct followers.
            for follower_name in self.system.followers:
                follower_class = self.system.follower_cls(follower_name)
                self.start_application(
                    follower_class, with_subprocess=with_subprocesses
                )

            # Construct non-follower leaders.
            for leader_name in self.system.leaders_only:
                leader_cls = self.system.leader_cls(leader_name)
                self.start_application(leader_cls, with_subprocess=with_subprocesses)

            # Construct singles.
            for name in self.system.singles:
                single_cls = self.system.get_app_cls(name)
                self.start_application(single_cls, with_subprocess=with_subprocesses)

            if with_subprocesses:
                # Wait for subprocesses to start.
                for server_subprocess in self.server_subprocesses:
                    server_subprocess.has_started.wait()
                self.has_errored.wait(timeout=1)
                for t in self.server_subprocesses:
                    if (not t.has_started.wait(timeout=0)) or t.has_errored.is_set():
                        self.stop()
                        raise Exception("Failed to start sub processes")
        finally:
            self.has_started.set()

    def start_application(
        self, app_class: Type[Application], with_subprocess: bool = False
    ) -> None:
        if self.has_errored.is_set():
            return
        if with_subprocess is False:
            self.servers[app_class.name] = start_server(
                app_class=app_class, env=self._env
            )

        else:
            self.server_subprocesses.append(
                start_server_subprocess(
                    app_class=app_class, env=self._env, has_errored=self.has_errored
                )
            )

    def get_client(
        self, cls: Type[TApplication]
    ) -> GrpcApplicationClient[TApplication]:
        try:
            client = self.clients[cls.name]
        except KeyError:
            # Construct application transcoder.
            # Make sure we don't try to connect to actual database.
            env = dict(self.env or {})
            env["PERSISTENCE_MODULE"] = "eventsourcing.popo"
            transcoder = cls(env=env).construct_transcoder()
            address = GrpcEnvironment(self._env).get_server_address(cls.name)
            client = GrpcApplicationClient(
                application_name=cls.name,
                owner_name=type(self).__name__,
                address=address,
                transcoder=transcoder,
                is_stopping=self.is_stopping,
            )
            client.connect()
            self.clients[cls.name] = client
        return cast(GrpcApplicationClient[TApplication], client)

    def get(self, cls: Type[TApplication]) -> TApplication:
        return self.get_client(cls).app

    def __del__(self) -> None:
        self.stop()

    def stop(self) -> None:
        try:
            with self.lock:
                if self.is_stopping.is_set():
                    return
                self.is_stopping.set()
            for client in self.clients.values():
                client.close()
            self.clients.clear()
            for server in self.servers.values():
                server.stop()
            for server in self.servers.values():
                server.has_stopped.wait()
            self.servers.clear()
            for process in self.server_subprocesses:
                process.stop()
            for process in self.server_subprocesses:
                process.has_stopped.wait()
            for app in self.apps.values():
                app.close()
            self.apps.clear()
        finally:
            self.has_stopped.set()



from __future__ import annotations

import os
import sys
import traceback
from itertools import count
from signal import SIGINT, SIGTERM, signal
from threading import Event, Lock
from typing import Any, cast

from eventsourcing.system import System
from eventsourcing.utils import TopicError, resolve_topic

from eventsourcing_grpc.application_server import start_server
from eventsourcing_grpc.runner import GrpcRunner


def eventsourcing_grpc_server() -> None:
    try:
        application_topic = os.environ["APPLICATION_TOPIC"]
        # print(f"Application topic found in env: {application_topic}")
    except KeyError:
        if len(sys.argv) > 1:
            application_topic = sys.argv[1]
            # print(f"Application topic found in args: {application_topic}")
        else:
            print("Error: application topic not set in env or args")
            raise SystemExit(1) from None

    is_interrupted = Event()
    lock = Lock()

    # Register signal handlers.
    def signal_handler(*args: Any) -> None:
        # print(
        #     f"Application server process received signal '{strsignal(signum)}'"
        # )
        with lock:
            if not is_interrupted.is_set():
                print(f"Stopping gRPC server: {application_topic}")
                is_interrupted.set()

    signal(SIGINT, signal_handler)
    signal(SIGTERM, signal_handler)

    try:
        print(
            f"Starting gRPC server: {application_topic}",
        )
        # sys.stdout.write(f"Starting subprocess {application_topic}\n")
        app_class = resolve_topic(application_topic)

        if "SYSTEM_TOPIC" in os.environ:
            system_topic = os.environ["SYSTEM_TOPIC"]
            system = cast(System, resolve_topic(system_topic))
            print(app_class.name, "system topic:", system_topic)
            # Make sure we have a leader class if leading.
            if app_class.name in system.leads:
                app_class = system.leader_cls(app_class.name)

        # Get the address and start the application server.
        server = start_server(
            app_class=app_class, env=os.environ, is_stopping=is_interrupted
        )

        # Wait for termination.
        server.wait_for_termination()
    except BaseException:
        print(traceback.format_exc())
        # print("Exiting with status code 1 after error")
        sys.exit(1)


def eventsourcing_grpc_runner() -> None:  # noqa: C901
    try:
        system_topic = os.environ["SYSTEM_TOPIC"]
        # print(f"System topic found in env: {system_topic}")
    except KeyError:
        if len(sys.argv) > 1:
            system_topic = sys.argv[1]
            # print(f"System topic found in args: {system_topic}")
        else:
            print("Error: system topic not set in env or args")
            raise SystemExit(1) from None

    is_interrupted = Event()
    lock = Lock()

    # Register signal handlers.
    def signal_handler(*args: Any) -> None:
        # print(
        #     f"Application server process received signal '{strsignal(signum)}'"
        # )
        with lock:
            if not is_interrupted.is_set():
                print(f"Stopping gRPC runner: {system_topic}")
                is_interrupted.set()

    signal(SIGINT, signal_handler)
    signal(SIGTERM, signal_handler)

    try:
        # Todo: If topic resolves to a module, search for system object in the module.
        try:
            system = cast(System, resolve_topic(system_topic))
        except TopicError:
            print(f"Error: system not found: {system_topic}")
            raise SystemExit(1) from None

        print(
            f"Starting gRPC runner: {system_topic}",
        )

        env = os.environ.copy()
        counter = count(50051)
        for app_name in system.nodes.keys():
            var_name = f"{app_name.upper()}_GRPC_SERVER_ADDRESS"
            port = next(counter)
            var_value = f"localhost:{port}"
            print(f"Setting {var_name}={var_value}")
            env[var_name] = var_value

        # Get the address and start the application server.
        with lock:
            if not is_interrupted.is_set():
                runner = GrpcRunner(system=system, env=env)

                runner.start(with_subprocesses=True)
                if not runner.has_started.wait(timeout=5):
                    is_interrupted.set()

        # Wait for termination.
        try:
            if runner.has_errored.is_set():
                runner.stop()
                runner.has_stopped.wait()
                raise RuntimeError("Couldn't start runner")
            is_interrupted.wait()
            runner.stop()
            runner.has_stopped.wait()
        except NameError:
            pass

    except SystemExit:
        raise
    except BaseException:
        try:
            runner.has_stopped.wait()
        except NameError:
            pass
        print(traceback.format_exc())
        # print("Exiting with status code 1 after error")
        sys.exit(1)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: eventsourcing_grpc/protos/application.proto
"""Generated protocol buffer code."""
from google.protobuf import (
    descriptor as _descriptor,
    descriptor_pool as _descriptor_pool,
    message as _message,
    reflection as _reflection,
    symbol_database as _symbol_database,
)

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n+eventsourcing_grpc/protos/application.proto\x12\x12\x65ventsourcing_grpc"\x07\n\x05\x45mpty"B\n\rMethodRequest\x12\x13\n\x0bmethod_name\x18\x01'
    b" \x01(\t\x12\x0c\n\x04\x61rgs\x18\x02 \x01(\x0c\x12\x0e\n\x06kwargs\x18\x03"
    b' \x01(\x0c"\x1b\n\x0bMethodReply\x12\x0c\n\x04\x64\x61ta\x18\x01'
    b' \x01(\x0c"R\n\x14NotificationsRequest\x12\r\n\x05start\x18\x01'
    b" \x01(\t\x12\r\n\x05limit\x18\x02 \x01(\t\x12\x0c\n\x04stop\x18\x03"
    b" \x01(\t\x12\x0e\n\x06topics\x18\x04"
    b' \x03(\t"k\n\x0cNotification\x12\n\n\x02id\x18\x01'
    b" \x01(\t\x12\x15\n\roriginator_id\x18\x02"
    b" \x01(\t\x12\x1a\n\x12originator_version\x18\x03 \x01(\t\x12\r\n\x05topic\x18\x04"
    b" \x01(\t\x12\r\n\x05state\x18\x05"
    b' \x01(\x0c"M\n\x12NotificationsReply\x12\x37\n\rnotifications\x18\x01'
    b" \x03(\x0b\x32"
    b' .eventsourcing_grpc.Notification"&\n\rPromptRequest\x12\x15\n\rupstream_name\x18\x01'
    b' \x01(\t2\xde\x02\n\x0b\x41pplication\x12>\n\x04Ping\x12\x19.eventsourcing_grpc.Empty\x1a\x19.eventsourcing_grpc.Empty"\x00\x12]\n\x15\x43\x61llApplicationMethod\x12!.eventsourcing_grpc.MethodRequest\x1a\x1f.eventsourcing_grpc.MethodReply"\x00\x12H\n\x06Prompt\x12!.eventsourcing_grpc.PromptRequest\x1a\x19.eventsourcing_grpc.Empty"\x00\x12\x66\n\x10GetNotifications\x12(.eventsourcing_grpc.NotificationsRequest\x1a&.eventsourcing_grpc.NotificationsReply"\x00\x42=\n!io.eventsourcing_grpc.applicationB\x10\x41pplicationProtoP\x01\xa2\x02\x03\x45SAb\x06proto3'
)


_EMPTY = DESCRIPTOR.message_types_by_name["Empty"]
_METHODREQUEST = DESCRIPTOR.message_types_by_name["MethodRequest"]
_METHODREPLY = DESCRIPTOR.message_types_by_name["MethodReply"]
_NOTIFICATIONSREQUEST = DESCRIPTOR.message_types_by_name["NotificationsRequest"]
_NOTIFICATION = DESCRIPTOR.message_types_by_name["Notification"]
_NOTIFICATIONSREPLY = DESCRIPTOR.message_types_by_name["NotificationsReply"]
_PROMPTREQUEST = DESCRIPTOR.message_types_by_name["PromptRequest"]
Empty = _reflection.GeneratedProtocolMessageType(
    "Empty",
    (_message.Message,),
    {
        "DESCRIPTOR": _EMPTY,
        "__module__": "eventsourcing_grpc.protos.application_pb2"
        # @@protoc_insertion_point(class_scope:eventsourcing_grpc.Empty)
    },
)
_sym_db.RegisterMessage(Empty)

MethodRequest = _reflection.GeneratedProtocolMessageType(
    "MethodRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _METHODREQUEST,
        "__module__": "eventsourcing_grpc.protos.application_pb2"
        # @@protoc_insertion_point(class_scope:eventsourcing_grpc.MethodRequest)
    },
)
_sym_db.RegisterMessage(MethodRequest)

MethodReply = _reflection.GeneratedProtocolMessageType(
    "MethodReply",
    (_message.Message,),
    {
        "DESCRIPTOR": _METHODREPLY,
        "__module__": "eventsourcing_grpc.protos.application_pb2"
        # @@protoc_insertion_point(class_scope:eventsourcing_grpc.MethodReply)
    },
)
_sym_db.RegisterMessage(MethodReply)

NotificationsRequest = _reflection.GeneratedProtocolMessageType(
    "NotificationsRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _NOTIFICATIONSREQUEST,
        "__module__": "eventsourcing_grpc.protos.application_pb2"
        # @@protoc_insertion_point(class_scope:eventsourcing_grpc.NotificationsRequest)
    },
)
_sym_db.RegisterMessage(NotificationsRequest)

Notification = _reflection.GeneratedProtocolMessageType(
    "Notification",
    (_message.Message,),
    {
        "DESCRIPTOR": _NOTIFICATION,
        "__module__": "eventsourcing_grpc.protos.application_pb2"
        # @@protoc_insertion_point(class_scope:eventsourcing_grpc.Notification)
    },
)
_sym_db.RegisterMessage(Notification)

NotificationsReply = _reflection.GeneratedProtocolMessageType(
    "NotificationsReply",
    (_message.Message,),
    {
        "DESCRIPTOR": _NOTIFICATIONSREPLY,
        "__module__": "eventsourcing_grpc.protos.application_pb2"
        # @@protoc_insertion_point(class_scope:eventsourcing_grpc.NotificationsReply)
    },
)
_sym_db.RegisterMessage(NotificationsReply)

PromptRequest = _reflection.GeneratedProtocolMessageType(
    "PromptRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _PROMPTREQUEST,
        "__module__": "eventsourcing_grpc.protos.application_pb2"
        # @@protoc_insertion_point(class_scope:eventsourcing_grpc.PromptRequest)
    },
)
_sym_db.RegisterMessage(PromptRequest)

_APPLICATION = DESCRIPTOR.services_by_name["Application"]
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = (
        b"\n!io.eventsourcing_grpc.applicationB\020ApplicationProtoP\001\242\002\003ESA"
    )
    _EMPTY._serialized_start = 67
    _EMPTY._serialized_end = 74
    _METHODREQUEST._serialized_start = 76
    _METHODREQUEST._serialized_end = 142
    _METHODREPLY._serialized_start = 144
    _METHODREPLY._serialized_end = 171
    _NOTIFICATIONSREQUEST._serialized_start = 173
    _NOTIFICATIONSREQUEST._serialized_end = 255
    _NOTIFICATION._serialized_start = 257
    _NOTIFICATION._serialized_end = 364
    _NOTIFICATIONSREPLY._serialized_start = 366
    _NOTIFICATIONSREPLY._serialized_end = 443
    _PROMPTREQUEST._serialized_start = 445
    _PROMPTREQUEST._serialized_end = 483
    _APPLICATION._serialized_start = 486
    _APPLICATION._serialized_end = 836
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from eventsourcing_grpc.protos import (
    application_pb2 as eventsourcing__grpc_dot_protos_dot_application__pb2,
)


class ApplicationStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Ping = channel.unary_unary(
            "/eventsourcing_grpc.Application/Ping",
            request_serializer=eventsourcing__grpc_dot_protos_dot_application__pb2.Empty.SerializeToString,
            response_deserializer=eventsourcing__grpc_dot_protos_dot_application__pb2.Empty.FromString,
        )
        self.CallApplicationMethod = channel.unary_unary(
            "/eventsourcing_grpc.Application/CallApplicationMethod",
            request_serializer=eventsourcing__grpc_dot_protos_dot_application__pb2.MethodRequest.SerializeToString,
            response_deserializer=eventsourcing__grpc_dot_protos_dot_application__pb2.MethodReply.FromString,
        )
        self.Prompt = channel.unary_unary(
            "/eventsourcing_grpc.Application/Prompt",
            request_serializer=eventsourcing__grpc_dot_protos_dot_application__pb2.PromptRequest.SerializeToString,
            response_deserializer=eventsourcing__grpc_dot_protos_dot_application__pb2.Empty.FromString,
        )
        self.GetNotifications = channel.unary_unary(
            "/eventsourcing_grpc.Application/GetNotifications",
            request_serializer=eventsourcing__grpc_dot_protos_dot_application__pb2.NotificationsRequest.SerializeToString,
            response_deserializer=eventsourcing__grpc_dot_protos_dot_application__pb2.NotificationsReply.FromString,
        )


class ApplicationServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Ping(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def CallApplicationMethod(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Prompt(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetNotifications(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_ApplicationServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "Ping": grpc.unary_unary_rpc_method_handler(
            servicer.Ping,
            request_deserializer=eventsourcing__grpc_dot_protos_dot_application__pb2.Empty.FromString,
            response_serializer=eventsourcing__grpc_dot_protos_dot_application__pb2.Empty.SerializeToString,
        ),
        "CallApplicationMethod": grpc.unary_unary_rpc_method_handler(
            servicer.CallApplicationMethod,
            request_deserializer=eventsourcing__grpc_dot_protos_dot_application__pb2.MethodRequest.FromString,
            response_serializer=eventsourcing__grpc_dot_protos_dot_application__pb2.MethodReply.SerializeToString,
        ),
        "Prompt": grpc.unary_unary_rpc_method_handler(
            servicer.Prompt,
            request_deserializer=eventsourcing__grpc_dot_protos_dot_application__pb2.PromptRequest.FromString,
            response_serializer=eventsourcing__grpc_dot_protos_dot_application__pb2.Empty.SerializeToString,
        ),
        "GetNotifications": grpc.unary_unary_rpc_method_handler(
            servicer.GetNotifications,
            request_deserializer=eventsourcing__grpc_dot_protos_dot_application__pb2.NotificationsRequest.FromString,
            response_serializer=eventsourcing__grpc_dot_protos_dot_application__pb2.NotificationsReply.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "eventsourcing_grpc.Application", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class Application(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Ping(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/eventsourcing_grpc.Application/Ping",
            eventsourcing__grpc_dot_protos_dot_application__pb2.Empty.SerializeToString,
            eventsourcing__grpc_dot_protos_dot_application__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def CallApplicationMethod(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/eventsourcing_grpc.Application/CallApplicationMethod",
            eventsourcing__grpc_dot_protos_dot_application__pb2.MethodRequest.SerializeToString,
            eventsourcing__grpc_dot_protos_dot_application__pb2.MethodReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Prompt(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/eventsourcing_grpc.Application/Prompt",
            eventsourcing__grpc_dot_protos_dot_application__pb2.PromptRequest.SerializeToString,
            eventsourcing__grpc_dot_protos_dot_application__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetNotifications(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/eventsourcing_grpc.Application/GetNotifications",
            eventsourcing__grpc_dot_protos_dot_application__pb2.NotificationsRequest.SerializeToString,
            eventsourcing__grpc_dot_protos_dot_application__pb2.NotificationsReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eventsourcing_grpc', 'eventsourcing_grpc.protos']

package_data = \
{'': ['*']}

install_requires = \
['eventsourcing>=9.2.11,<9.3.0',
 'grpcio>=1.44.0,<1.45.0',
 'protobuf>=3.20.0,<3.21.0']

entry_points = \
{'console_scripts': ['eventsourcing_grpc_runner = '
                     'eventsourcing_grpc.console:eventsourcing_grpc_runner',
                     'eventsourcing_grpc_server = '
                     'eventsourcing_grpc.console:eventsourcing_grpc_server']}

setup_kwargs = {
    'name': 'eventsourcing-grpc',
    'version': '0.1.2',
    'description': 'Python package for eventsourcing with gRPC',
    'long_description': '# Event Sourcing with gRPC\n\nThis is an extension for [the Python eventsourcing library](https://github.com/pyeventsourcing/eventsourcing).\n\nThis package provides a gRPC server class that runs event-sourced applications\nin a gRPC server. A gRPC client class can be used to interact with the application\nserver. Application command and query methods can be called, and\nnotifications selected, remotely using an application proxy object\nprovided by the client.\n\nThis package also provides a system runner class, with the same interface as the Runner\nclasses in the core eventsourcing library. It can run a system of event-sourced\napplications as a set of gRPC servers. Leader and follower connections, for prompting\nand pulling event notifications, are implemented with gRPC clients.\n\nThis package provides console commands that can be used to start servers and run systems\nfrom a command line interface.\n\nThis package also provides an example of running an event-sourced application in\ngRPC server in a Docker container.\n\nThis package also provides an example of running a system of applications\nin gRPC servers in Docker containers with docker-compose.\n\nThis package also provides an example of running a system of applications\nin gRPC servers in Docker containers with Kubernetes, with clients and\nservers authenticated with SSL/TLS PKI certificates.\n\n## Installation\n\nUse pip to install the [stable distribution](https://pypi.org/project/eventsourcing-grpc/)\nfrom the Python Package Index.\n\n    $ pip install eventsourcing_grpc\n\nIt is recommended to install Python packages into a Python virtual environment.\n\n\n## gRPC clients and servers\n\nThe `GrpcApplicationServer` class can be used to run an event-sourced\napplication as a gRPC server, and the `GrpcApplicationClient` class can\nbe used to connect to a server.\n\nDefine an event-sourced application.\n\n```python\nfrom eventsourcing_grpc.example import Orders\n```\n\nConfigure with environment variables.\n\n```python\nimport os\n\nos.environ["ORDERS_GRPC_SERVER_ADDRESS"] = "localhost:50051"\n```\n\nApplication persistence can be configured in the usual way (see core library docs).\n\nUse the `start_server()` function to create and start a gRPC application server.\n\n```python\nfrom eventsourcing_grpc.application_server import start_server\n\nserver = start_server(Orders)\n```\n\nUse the `connect()` function to create a gRPC client and connect to the server.\n\n```python\nfrom eventsourcing_grpc.application_client import connect\n\nclient = connect(Orders, max_attempts=10)\n```\n\nUse the application proxy to call command and query methods on the application server.\n\n```python\n# Create order.\norder_id = client.app.create_new_order()\n\n# Get order.\norder = client.app.get_order(order_id)\nassert order["id"] == order_id\nassert order["is_reserved"] == False\nassert order["is_paid"] == False\n```\n\nUse the application proxy to select event notifications.\n\n```python\nnotifications = client.app.notification_log.select(start=1, limit=10)\nassert len(notifications) == 1\nassert notifications[0].id == 1\nassert notifications[0].originator_id == order_id\nassert notifications[0].originator_version == 1\nassert notifications[0].topic == "eventsourcing_grpc.example:Order.Created"\n```\n\nClose client and stop server.\n\n```python\nclient.close()\nserver.stop(wait=True)\n```\n\n\n## gRPC runner\n\nThe `GrpcRunner` class can be used to run a system of event-sourced applications\nas an inter-connected set of gRPC servers.\n\n```python\nfrom eventsourcing_grpc.runner import GrpcRunner\n```\n\nDefine a system of event-sourced applications.\n\n```python\nfrom eventsourcing_grpc.example import system\n```\n\nConfigure with environment variables.\n\n```python\nimport os\n\nos.environ["ORDERS_GRPC_SERVER_ADDRESS"] = "localhost:50051"\nos.environ["RESERVATIONS_GRPC_SERVER_ADDRESS"] = "localhost:50052"\nos.environ["PAYMENTS_GRPC_SERVER_ADDRESS"] = "localhost:50053"\n```\n\nCreate and start a gRPC system runner (starts servers and connects clients).\n\n```python\n\nrunner = GrpcRunner(system)\nrunner.start(with_subprocesses=True)\n```\n\nGet an application proxy from the runner.\n\n```python\napp = runner.get(Orders)\n```\n\nCall command and query methods on application proxy.\n\n```python\norder1_id = app.create_new_order()\n\n# Wait for the processing to happen.\nfrom time import sleep\n\nfor _ in range(100):\n    sleep(0.1)\n    if app.is_order_paid(order1_id):\n        break\n    elif runner.has_errored.is_set():\n        raise AssertionError("Runner error")\nelse:\n    raise AssertionError("Timeout waiting for order to be paid")\n\n```\n\nGet application\'s event notifications.\n\n```python\nnotifications = app.notification_log.select(start=1, limit=10)\nassert len(notifications) == 3\nassert notifications[0].id == 1\nassert notifications[0].originator_id == order1_id\nassert notifications[0].originator_version == 1\nassert notifications[0].topic == "eventsourcing_grpc.example:Order.Created"\nassert notifications[1].id == 2\nassert notifications[1].originator_id == order1_id\nassert notifications[1].originator_version == 2\nassert notifications[1].topic == "eventsourcing_grpc.example:Order.Reserved"\nassert notifications[2].id == 3\nassert notifications[2].originator_id == order1_id\nassert notifications[2].originator_version == 3\nassert notifications[2].topic == "eventsourcing_grpc.example:Order.Paid"\n\n```\n\nStop runner (closes clients and stops servers).\n\n```python\nrunner.stop()\n```\n\n## Start server from command line\n\nThe console command `eventsourcing_grpc_server` can be used to start\na gRPC application server. The application topic can be given as a\ncommand line argument.\n\n    $ export ORDERS_GRPC_SERVER_ADDRESS=localhost:50051\n    $ eventsourcing_grpc_server eventsourcing_grpc.example:Orders\n\nA system of interconnected application servers can be started in this way,\nafter setting `SYSTEM_TOPIC` in the environment.\n\n    $ export SYSTEM_TOPIC=eventsourcing_grpc.example:system\n    $ export ORDERS_GRPC_SERVER_ADDRESS=localhost:50051\n    $ export RESERVATIONS_GRPC_SERVER_ADDRESS=localhost:50052\n    $ export PAYMENTS_GRPC_SERVER_ADDRESS=localhost:50053\n\n    $ eventsourcing_grpc_server eventsourcing_grpc.example:Orders &\n    $ eventsourcing_grpc_server eventsourcing_grpc.example:Reservations &\n    $ eventsourcing_grpc_server eventsourcing_grpc.example:Payments &\n\nIn the example above, the servers are run as background processes.\n\n## Start runner from command line\n\nThe console command `eventsourcing_grpc_runner` can be used to run a system\nof event-sourced applications. Only the system topic must be configured.\n\nThe system topic can be given as a command line argument.\n\n    $ eventsourcing_grpc_runner eventsourcing_grpc.example:system\n\nAlternatively, the system topic can be set using the `SYSTEM_TOPIC` environment variable.\n\n    $ export SYSTEM_TOPIC=eventsourcing_grpc.example:system\n    $ eventsourcing_grpc_runner\n\nThe server addresses will be automatically configured, unless they are set as\nenvironment variables.\n\n## Docker container\n\nA Docker file can be defined to run the `eventsourcing_grpc_server` command\nmentioned above, after installing an event-sourced system and the `eventsourcing_grpc`\npackage.\n\n```dockerfile\nFROM python:3.10\n\nWORKDIR /app\n\n# Install package(s).\nRUN pip install eventsourcing_grpc\n\n# Run application server.\nENV PYTHONUNBUFFERED = 1\nCMD ["eventsourcing_grpc_server"]\n```\n\nA container image can then be built with Docker.\n\n    $ docker build -t eventsourcing-grpc -f Dockerfile .\n\n## Docker Compose\n\nA system of gRPC application servers can be run with Docker Compose.\nThe following Docker Compose file assumes SSL/TLS PKI certificates\nhave been generated in a local `./ssl` folder.\n\n```\nversion: \'2\'\n\nservices:\n  orders:\n    image: "eventsourcing-grpc:v1"\n    environment:\n      APPLICATION_TOPIC: "eventsourcing_grpc.example:Orders"\n      SYSTEM_TOPIC: "eventsourcing_grpc.example:system"\n      ORDERS_GRPC_SERVER_ADDRESS: "orders:50051"\n      RESERVATIONS_GRPC_SERVER_ADDRESS: "reservations:50052"\n      PAYMENTS_GRPC_SERVER_ADDRESS: "payments:50053"\n      MAX_PULL_INTERVAL: "10"\n      GRPC_SSL_PRIVATE_KEY_PATH: /app/ssl/orders.key\n      GRPC_SSL_CERTIFICATE_PATH: /app/ssl/orders.crt\n      GRPC_SSL_ROOT_CERTIFICATE_PATH: /app/ssl/root.crt\n    volumes:\n      - ./ssl/orders:/app/ssl\n    ports:\n      - "50051:50051"\n\n  reservations:\n    image: "eventsourcing-grpc:v1"\n    environment:\n      APPLICATION_TOPIC: "eventsourcing_grpc.example:Reservations"\n      SYSTEM_TOPIC: "eventsourcing_grpc.example:system"\n      ORDERS_GRPC_SERVER_ADDRESS: "orders:50051"\n      RESERVATIONS_GRPC_SERVER_ADDRESS: "reservations:50052"\n      PAYMENTS_GRPC_SERVER_ADDRESS: "payments:50053"\n      MAX_PULL_INTERVAL: "10"\n      GRPC_SSL_PRIVATE_KEY_PATH: /app/ssl/reservations.key\n      GRPC_SSL_CERTIFICATE_PATH: /app/ssl/reservations.crt\n      GRPC_SSL_ROOT_CERTIFICATE_PATH: /app/ssl/root.crt\n    volumes:\n      - ./ssl/reservations:/app/ssl\n    ports:\n      - "50052:50052"\n\n  payments:\n    image: "eventsourcing-grpc:v1"\n    environment:\n      APPLICATION_TOPIC: "eventsourcing_grpc.example:Payments"\n      SYSTEM_TOPIC: "eventsourcing_grpc.example:system"\n      ORDERS_GRPC_SERVER_ADDRESS: "orders:50051"\n      RESERVATIONS_GRPC_SERVER_ADDRESS: "reservations:50052"\n      PAYMENTS_GRPC_SERVER_ADDRESS: "payments:50053"\n      MAX_PULL_INTERVAL: "10"\n      GRPC_SSL_PRIVATE_KEY_PATH: /app/ssl/payments.key\n      GRPC_SSL_CERTIFICATE_PATH: /app/ssl/payments.crt\n      GRPC_SSL_ROOT_CERTIFICATE_PATH: /app/ssl/root.crt\n    volumes:\n      - ./ssl/payments:/app/ssl\n    ports:\n      - "50053:50053"\n```\n\nThe containers can then be started.\n\n    $ docker-compose up -d\n\n## Kubernetes\n\nA system of gRPC application servers can be run with Kubernetes.\n\nCreate Kubernetes Secrets from the SSL/TLS PKI certificates.\n\n\t$ kubectl create secret generic root-ssl-secret \\\n    --from-file=root.crt=ssl/root/root.crt\n\n\t$ kubectl create secret tls orders-ssl-secret \\\n\t--key ./ssl/orders/orders.key \\\n    --cert ./ssl/orders/orders.crt\n\n\t$ kubectl create secret tls reservations-ssl-secret \\\n\t--key ./ssl/reservations/reservations.key \\\n    --cert ./ssl/reservations/reservations.crt\n\n\t$ kubectl create secret tls payments-ssl-secret \\\n\t--key ./ssl/payments/payments.key \\\n    --cert ./ssl/payments/payments.crt\n\n\nCreate a Kubernetes Deployment for each application.\n\n```yaml\napiVersion: apps/v1\nkind: Deployment\nmetadata:\n  name: orders\n  labels:\n    system: orders\nspec:\n  selector:\n    matchLabels:\n      app: orders\n  replicas: 1\n  template:\n    metadata:\n      labels:\n        app: orders\n    spec:\n      volumes:\n      - name: ssl-secret-volume\n        secret:\n          secretName: orders-ssl-secret\n      - name: root-ssl-secret-volume\n        secret:\n          secretName: root-ssl-secret\n      containers:\n      - name: orders\n        image: "eventsourcing-grpc:latest"\n        imagePullPolicy: Never\n        ports:\n        - containerPort: 50051\n        volumeMounts:\n          - mountPath: /app/ssl\n            name: ssl-secret-volume\n          - mountPath: /app/root-ssl\n            name: root-ssl-secret-volume\n        env:\n          - name: APPLICATION_TOPIC\n            value: "eventsourcing_grpc.example:Orders"\n          - name: SYSTEM_TOPIC\n            value: "eventsourcing_grpc.example:system"\n          - name: ORDERS_GRPC_SERVER_ADDRESS\n            value: "0.0.0.0:50051"\n          - name: RESERVATIONS_GRPC_SERVER_ADDRESS\n            value: "reservations:50052"\n          - name: PAYMENTS_GRPC_SERVER_ADDRESS\n            value: "payments:50053"\n          - name: GRPC_SSL_PRIVATE_KEY_PATH\n            value: /app/ssl/tls.key\n          - name: GRPC_SSL_CERTIFICATE_PATH\n            value: /app/ssl/tls.crt\n          - name: GRPC_SSL_ROOT_CERTIFICATE_PATH\n            value: /app/root-ssl/root.crt\n```\n\nCreate a Kubernetes Service for each Deployment.\n\n```yaml\napiVersion: v1\nkind: Service\nmetadata:\n  name: orders\n  labels:\n    system: orders\nspec:\n  selector:\n    app: orders\n  type: LoadBalancer\n  ports:\n  - port: 50051\n    targetPort: 50051\n    protocol: TCP\n```\n\nThe above configuration will run each application on a separate Pod. Alternatively,\nyou can run all the applications in the same Pod, by defining a multi-container\nDeployment.\n\nPlease refer to the project repo (and its GitHub actions workflow and Makefile)\nto see a fully working example of running a system of applications with minikube.\n\n## Developers\n\nAfter cloning the eventsourcing-grpc repository, set up a virtual\nenvironment and install dependencies by running the following command in the\nroot folder.\n\n    $ make install\n\nThe ``make install`` command uses the build tool Poetry to create a dedicated\nPython virtual environment for this project, and installs popular development\ndependencies such as Black, isort and pytest.\n\nAdd tests in `./tests`. Add code in `./eventsourcing_grpc`.\n\nGenerate SSL/TLS certificates and private keys for testing.\n\n    $ make ssl\n\nRun tests.\n\n    $ make test\n\nCheck the formatting of the code.\n\n    $ make lint\n\nReformat the code.\n\n    $ make fmt\n\nAdd dependencies in `pyproject.toml` and then update installed packages.\n\n    $ make update-packages\n\nIf you edit the .proto files, regenerate the project\'s protos package.\n\n    $ make grpc-stubs\n',
    'author': 'John Bywater',
    'author_email': 'john.bywater@appropriatesoftware.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pyeventsourcing/eventsourcing-grpc',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Optional, cast

from eventsourcing.system import System
from eventsourcing.utils import Environment, EnvType, resolve_topic


class GrpcEnvironment:
    def __init__(self, env: EnvType):
        self.env = env

    def get_server_address(self, name: str) -> str:
        env = Environment(name=name, env=self.env)
        address = env.get("GRPC_SERVER_ADDRESS")
        if not address:
            print(f"{name} gRPC server address not found in environment")
            raise ValueError(f"{name} gRPC server address not found in environment")
        else:
            return address

    def get_max_pull_interval(self, name: str) -> float:
        env = Environment(name=name, env=self.env)
        max_pull_interval_str = env.get("MAX_PULL_INTERVAL")
        if not max_pull_interval_str:
            return 10
        try:
            max_pull_interval = float(max_pull_interval_str)
        except ValueError:
            raise ValueError(
                "Could not covert MAX_PULL_INTERVAL to float: {max_pull_interval_str}"
            ) from None
        else:
            return max_pull_interval

    def get_system(self) -> Optional[System]:
        system_topic = self.env.get("SYSTEM_TOPIC")
        if system_topic:
            return cast(System, resolve_topic(system_topic))
        else:
            return None

    def get_ssl_root_certificate_path(self) -> Optional[str]:
        return self.env.get("GRPC_SSL_ROOT_CERTIFICATE_PATH")

    def get_ssl_certificate_path(self) -> Optional[str]:
        return self.env.get("GRPC_SSL_CERTIFICATE_PATH")

    def get_ssl_private_key_path(self) -> Optional[str]:
        return self.env.get("GRPC_SSL_PRIVATE_KEY_PATH")
